/*
 * libspe - A wrapper library to adapt the JSRE SPE usage model to SPUFS
 * Copyright (C) 2005 IBM Corp.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 *  License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <elf.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <spe.h>
#include <elf_loader.h>
#include <mfc.h>
#include <spe_exec.h>

#define __PRINTF(fmt, args...) { fprintf(stderr,fmt , ## args); }
#ifdef DEBUG
#define DEBUG_PRINTF(fmt, args...) __PRINTF(fmt , ## args)
#else
#define DEBUG_PRINTF(fmt, args...)
#endif

/*
 * Helpers
 *
 * */

struct thread_store
{
	pthread_t spe_thread;
	int fd_mem, fd_run, fd_mbox, fd_mbox_stat, fd_ibox, fd_ibox_stat, fd_wbox, fd_wbox_stat, fd_npc, fd_sig1, fd_sig2;
	int policy, priority, use_events, affinity_mask, spe_flags ;
	int npc, ret_status;
	spe_gid_t group_id;
	void *mem_mmap_base;
};

struct group_store
{
	int policy, priority, use_events;
};

/*Default SPE group options
 */
static struct group_store default_group = {SCHED_OTHER, 0, 0 };

/*
 * int spe_ldr[]:
 * SPE code that performs the actual parameter setting:
 */
static int spe_ldr[] = {
	0x30fff203,
	0x30fff404,
	0x30fff605,
	0x30fff806,
	0x30fff000,
	0x35000000,
	0x00000000,
	0x00000000
};

/*
 * int spe_ldr_debug[]:
 * Debug variant of loader that executes a breakpoint instruction
 * before starting main SPE code:
 */
static int spe_ldr_debug[] = {
	0x30fff203,
	0x30fff404,
	0x30fff605,
	0x30fff806,
	0x30fff000,
	0x00003fff,
	0x35000000,
	0x00000000
};

/**
 * Low-level API
 * 
 */

static int
spe_open_files(struct thread_store *spe, const char *pathname)
{
	char filename[40];
	
	sprintf(filename, "%s/run", pathname);
	spe->fd_run = open(filename, O_RDONLY);
	
	sprintf(filename, "%s/mbox", pathname);
	spe->fd_mbox = open(filename, O_RDONLY);

	sprintf(filename, "%s/mbox_stat", pathname);
	spe->fd_mbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox", pathname);
	spe->fd_ibox = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox_stat", pathname);
	spe->fd_ibox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/wbox", pathname);
	spe->fd_wbox = open(filename, O_WRONLY);

	sprintf(filename, "%s/wbox_stat", pathname);
	spe->fd_wbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/npc", pathname);
	spe->fd_npc = open(filename, O_RDWR);
		
	sprintf(filename, "%s/signal1", pathname);
	spe->fd_sig1 = open(filename, O_WRONLY);
		
	sprintf(filename, "%s/signal2", pathname);
	spe->fd_sig2 = open(filename, O_WRONLY);
		
	if (spe->fd_mbox      < 0 || spe->fd_ibox      < 0 || spe->fd_wbox      < 0 ||
	    spe->fd_mbox_stat < 0 || spe->fd_ibox_stat < 0 || spe->fd_wbox_stat < 0 ||
	    spe->fd_npc       < 0 || spe->fd_sig1      < 0 || spe->fd_sig2      < 0) {
		DEBUG_PRINTF("Could not open all SPE files.\n");
		close(spe->fd_mbox);
		close(spe->fd_mbox_stat);
		close(spe->fd_ibox);
		close(spe->fd_ibox_stat);
		close(spe->fd_wbox);
		close(spe->fd_wbox_stat);
		close(spe->fd_npc);
		close(spe->fd_sig1);
		close(spe->fd_sig2);
		return -1;
	}
	return 0;
}

void *
spe_gid_setup(int policy, int priority, int use_events)
{
	struct group_store *group_store;

        group_store = calloc (1, sizeof *group_store);
	if (!group_store)
	{
		DEBUG_PRINTF ("Could not allocate group store\n");
		errno = ENOMEM;
	        return NULL;
	}
	
	group_store->policy = policy;
	group_store->priority = priority;
	group_store->use_events = use_events;
	
	return group_store;
}

void *
spe_setup (spe_gid_t gid, spe_program_handle_t *handle, void *argp, void *envp, int flags)
{
	struct thread_store *thread_store;
	int rc, memfd, sigfd;
	addr64 argp64, envp64;
	char signame[40], memname[40], pathname[40];
	void *spe_ld_buf;
#ifdef SPE_NO_MMAP
	ssize_t count = 0, num = 0;
#endif
	struct spe_ld_info ld_info;
	struct spe_exec_params spe_params __attribute__ ((aligned (4096)));
	
	if(!gid)
	{
		gid=&default_group;
	}

	/* Allocate thread structure */

	thread_store = calloc (1, sizeof *thread_store);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("Could not allocate thread store\n");
		  errno = ENOMEM;
		  return NULL;
	  }

	argp64.ull = (unsigned long long) (unsigned long) argp;
	envp64.ull = (unsigned long long) (unsigned long) envp;

	thread_store->group_id = gid;
	
	/* Make the SPE Directory */

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	DEBUG_PRINTF ("create %s\n", pathname);

	rc = mkdir(pathname, S_IRUSR | S_IWUSR | S_IXUSR);
	
	if (rc < 0)
	  {
		  DEBUG_PRINTF ("Could not create SPE %s\n", pathname);
		  return NULL;
	  }

	sprintf (memname, "%s/mem", pathname);

	/* Check SPE */

	memfd = open (memname, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	if (memfd < 0)
	  {
		  DEBUG_PRINTF ("Could not open SPE mem file.\n");
		  return NULL;
	  }

	/* Prepare Loader */
#ifdef SPE_NO_MMAP	
	spe_ld_buf = malloc (LS_SIZE);
#else
	spe_ld_buf = mmap(0, LS_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, 0);
	thread_store->mem_mmap_base = spe_ld_buf;
	thread_store->fd_mem = memfd;
#endif
	if (!spe_ld_buf)
	  {
		  DEBUG_PRINTF ("Could not allocate SPE memory. \n");
		  errno = ENOMEM;
		  return NULL;
	  }

	memset(spe_ld_buf, 0, LS_SIZE);

	rc = load_spe_elf (handle, spe_ld_buf, &ld_info);
	if (rc != 0)
	  {
		  DEBUG_PRINTF ("Load SPE ELF failed..\n");
		  return NULL;
	  }

	/* Add SPE exec program */

	if (!(flags & SPE_PTRACE_ME))
	  {
		  DEBUG_PRINTF ("Add exec prog dst:0x%04x size:0x%04x\n",
			        SPE_LDR_START, sizeof (spe_ldr));
		  memcpy (spe_ld_buf + SPE_LDR_START, &spe_ldr,
			  sizeof (spe_ldr));
	  }
	else
	  {
		  DEBUG_PRINTF ("Add debug exec prog dst:0x%04x size:0x%04x\n",
			        SPE_LDR_START, sizeof (spe_ldr_debug));
		  memcpy (spe_ld_buf + SPE_LDR_START, &spe_ldr_debug,
			  sizeof (spe_ldr_debug));
	  }

	/* Add SPE exec parameters */

	spe_params.gpr0[0] = ld_info.entry;
	spe_params.gpr0[1] = 0;
	spe_params.gpr0[2] = 0;
	spe_params.gpr0[3] = 0;
	
	if (flags & SPE_USER_REGS) {
		/* When flags & SPE_USER_REGS is set, argp points
		 * to an array of 3x128b registers to be passed
		 * directly to the SPE program.
		 */
		memcpy(&spe_params.gpr3[0], argp, sizeof(unsigned int) * 12);
	} else {
		/* Regular parameter convention for spe_create_thread. */
		spe_params.gpr3[0] = 0;
		spe_params.gpr3[1] = 0;
		spe_params.gpr3[2] = 0;
		spe_params.gpr3[3] = 0;
	
		spe_params.gpr4[0] = argp64.ui[0];
		spe_params.gpr4[1] = argp64.ui[1];
		spe_params.gpr4[2] = 0;
		spe_params.gpr4[3] = 0;
	
		spe_params.gpr5[0] = envp64.ui[0];
		spe_params.gpr5[1] = envp64.ui[1];
		spe_params.gpr5[2] = 0;
		spe_params.gpr5[3] = 0;
	}

	spe_params.gpr6[0] = 0;
	spe_params.gpr6[1] = 0;
	spe_params.gpr6[2] = 0;
	spe_params.gpr6[3] = 0;

	DEBUG_PRINTF ("Add exec param dst:0x%04x size:0x%04x\n",
		      SPE_PARAM_START, sizeof (spe_params));
	memcpy (spe_ld_buf + SPE_PARAM_START, &spe_params,
		sizeof (spe_params));

#ifdef SPE_NO_MMAP
	/* Copy SPE image to SPUfs */
	do
	  {
		  num = write (memfd, spe_ld_buf + count, LS_SIZE - count);
		  if (num == -1)
		    {
			    DEBUG_PRINTF ("Transfer SPE ELF failed..\n");
			    return NULL;
		    }

		  count += num;
	  }
	while (count < LS_SIZE && num);
	close (memfd);

	/* Free the SPE Buffer */
	free (spe_ld_buf);
#endif

	rc = spe_open_files(thread_store, pathname);
	if (rc) {
		perror("open files\n");
		return NULL;
	}

	/*ToDo: Set flags */
	if ( flags & SPE_CFG_SIGNOTIFY1_OR )
	{
		char one[] = "1";
		
        	sprintf (signame, "%s/signal1_type", pathname);
        	sigfd = open (signame, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
        	if (sigfd < 0)
        	{
         		DEBUG_PRINTF ("Could not open SPE mem file.\n");
         		return NULL;
        	}
		write (sigfd, one, 1);
		close(sigfd);
	}
	if ( flags & SPE_CFG_SIGNOTIFY2_OR )
	{
		char one[] = "1";
		
        	sprintf (signame, "%s/signal2_type", pathname);
        	sigfd = open (signame, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
        	if (sigfd < 0)
        	{
         		DEBUG_PRINTF ("Could not open SPE mem file.\n");
         		return NULL;
        	}
		write (sigfd, one, 1);
		close(sigfd);
	}

	sync();

	return thread_store;
}

void
spe_cleanup (void *ptr)
{
	struct thread_store *thread_store = ptr;
	char pathname[40];

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	close (thread_store->fd_run);
	close (thread_store->fd_mbox);
	close (thread_store->fd_mbox_stat);
	close (thread_store->fd_ibox);
	close (thread_store->fd_ibox_stat);
	close (thread_store->fd_wbox);
	close (thread_store->fd_wbox_stat);
	close (thread_store->fd_npc);
	close (thread_store->fd_sig1);
	close (thread_store->fd_sig2);

#ifndef SPE_NO_MMAP
	munmap(thread_store->mem_mmap_base,LS_SIZE);
	close (thread_store->fd_mem);
#endif

#ifndef NODIRREMOVE
	rmdir (pathname);
#endif

	free (thread_store);
}

unsigned int
set_npc (void *ptr, unsigned int npc)
{
	struct thread_store *thread_store = ptr;
	int ret;
	char npcText[9];

	sprintf(npcText,"0x%06x\n",npc);
	ret = write(thread_store->fd_npc, npcText, 9);
	if (ret < 0)
        {
                return -EINVAL;
        }
	return npc;
}

int
do_spe_run (void *ptr)
{
	struct thread_store *thread_store = ptr;
	int runfd;

	struct spufs_run_arg
	{
		unsigned int npc;	/* inout: Next Program Counter */
		unsigned short code;	/* out:   SPE status */
		unsigned short status;
	};
	struct spufs_run_arg arg = { SPE_LDR_PROG_start, 0, 0 };
	int ret;

        DEBUG_PRINTF ("do_spe_run\n");
	runfd=thread_store->fd_run;

	ret = set_npc(thread_store, arg.npc);
	if (ret < 0)
	{
		DEBUG_PRINTF ("Could not set NPC in NPC file.\n");
		return -EINVAL;
	}

	ret = read(runfd, &arg, 8);
              
	DEBUG_PRINTF ("SPE (npc=0x%06x) (code=0x%04x) (status=0x%04x).\n",
		      arg.npc, arg.code, arg.status);
	if ( ret < 0) {
       		DEBUG_PRINTF ("Could not read() on SPE run file.\n");
		return -EINVAL;
	}


	DEBUG_PRINTF ("SPE thread result: %08x:%04x:%04x\n", arg.npc,
		      arg.code, arg.status);
	
	/* Save status & npc */	
	thread_store->npc = arg.npc;
	thread_store->ret_status = arg.code;

	return arg.code;
}

/**
 * \defgroup Interface
 */
/*@{*/

/* APIs for loading SPE-ELF images & files.
 */

/**
 * \brief spe_open_image - maps spe executable int system memory.
 * 
 * maps a spe executable into system memory and returns
 * a spe_program_handle for use with spe_create_thread.
 * 
 * @param filename file name of the spe executable to map
 * \returns handle of the program
 * \retval NULL failure, see \c errno for more info
 */
spe_program_handle_t 
*spe_open_image(const char *filename)
{
	spe_program_handle_t *ret;
	int binfd, f_stat;
	struct stat statbuf;
	
	ret=malloc(sizeof(spe_program_handle_t));
	if (!ret)
		return NULL;

	binfd = open(filename, O_RDONLY);
	if (binfd < 0)
		return NULL;
        
	f_stat = fstat(binfd, &statbuf);
	if (f_stat < 0)
		return NULL;
	
	ret->toe_struct = NULL;
	ret->prog_size = statbuf.st_size;

	ret->start = mmap(NULL, statbuf.st_size, PROT_READ , MAP_SHARED, binfd, 0);
	if (ret->start == MAP_FAILED)
		return NULL;

	return ret;
}

/**
 * \brief spe_close_image - unmaps spe executable from system memory.
 * 
 * Removes memory mapped file from system memory and releases allocated storage.
 * 
 * @param program_handle: hanldle of the spe executable
 * \retval zero on success
 * \retval -1 on failure, and sets \c errno.
 */
int 
spe_close_image(spe_program_handle_t *program_handle)
{
	void *start = program_handle->start;
	unsigned int prog_size = program_handle->prog_size;
	
	free(program_handle);

	return munmap(start, prog_size);
}

static void *
spe_thread (void *ptr)
{
	return (void *) (unsigned long) do_spe_run (ptr);
}

/**
 * Library API
 * 
 */

/**
 * \brief spe_create_group - allocates a new spe thread group.
 * 
 * Creates a spe thread group with given attributes.
 * 
 * @param policy scheduling class for spe threads
 * @param priority spe groups scheduling priority
 * @param use_events RFU
 * \returns group store structure
 * \retval NULL on failure, and sets \c errno.
 */
spe_gid_t
spe_create_group (int policy, int priority, int use_events)
{
	struct group_store *group_store;

	DEBUG_PRINTF ("spe_create_group(0x%x, 0x%x, 0x%x)\n",
			                  policy, priority, use_events);

	/*ToDo: Sanity check for priority versus policy*/
	
	group_store=spe_gid_setup(policy, priority, use_events);
	DEBUG_PRINTF("gid is %p\n", group_store);

	return group_store;
}

/**
 * \brief spe_group_max - returns maximum number of spe threads in this group
 *
 * @param spe_gid spe_gid group store structure
 * \return  maximum number of threads in this group.
 */
int
spe_group_max (spe_gid_t spe_gid)
{
	DEBUG_PRINTF ("spe_group_max(0x%x)\n", spe_gid);

	return SPE_MAX_THREADS_PER_GROUP;
}


/**
 *  \brief spe_create_thread - creates a spe thread
 * 
 * creates a spe thread of control that can be executed separately 
 * of the calling task.
 * 
 * @param gid spe_gid group store structure
 * @param handle a handle to a spe executable
 * @param argp ppe pointer to application specific data
 * @param envp ppe pointer to environment specific data
 * @param mask RFU
 * @param flags modifiers that are applied when the new thread is started
 * \returns spe thread store structure
 * \retval NULL on failure, and sets \c errno
 */
speid_t
spe_create_thread (spe_gid_t gid, spe_program_handle_t *handle, void *argp, void *envp, unsigned long mask,
		   int flags)
{
	struct thread_store *thread_store;
	int rc;

	DEBUG_PRINTF ("spe_create_thread(%p, %p, %p, %p, 0x%x, 0x%x)\n",
		      gid, handle, argp, envp, mask, flags);

	thread_store = spe_setup (gid, handle, argp, envp, flags);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("spe_setup failed\n");
		  errno=EINVAL;
		  return NULL;
	  }
	
	/*ToDo: set affinity mask, and priority */
	
	rc = pthread_create (&thread_store->spe_thread, NULL,
			     spe_thread, thread_store);
	if (rc)
	  {
		  perror ("pthread_create");
		  return NULL;
	  }

	return thread_store;
}

/**
 * \brief spe_wait - waits for spe thread to end
 *
 * spe_wait suspende execution of the current process until the specified
 * spe thread has exited.
 * 
 * @param speid spe store structure
 * @param status the spe threads return value is stored here
 * @param options RFU
 * 
 * \returns thread exit code.
 */
int
spe_wait (speid_t speid, int *status, int options)
{
	struct thread_store *thread_store = speid;
	void *pthread_rc;
	int rc;

	DEBUG_PRINTF ("spe_wait(0x%x, %p, 0x%x)\n", speid, status, options);

	rc = pthread_join (thread_store->spe_thread, &pthread_rc);
	*status = (int) (unsigned long) pthread_rc;
	
	/*ToDo: return code*/
	
	spe_cleanup (thread_store);

	DEBUG_PRINTF ("Thread ended.\n");
	return rc;
}

/**
 * \brief spe_kill - sends a signal to a spe thread
 *
 * spe_kill sends a signal to the specified process.
 * 
 * @param speid spe store structure
 * @param sig signal to send
 * 
 * \retval zero on success
 * \retval -1 on failure.
 */
int
spe_kill (speid_t speid, int sig)
{
	struct thread_store *thread_store = speid;
	int rc;

	thread_store=speid;	

	rc = pthread_kill (thread_store->spe_thread, sig);

	return rc;
}

/**
 * \brief spe_get_ls - returns the address of the local store in system memory
 * 
 * spe_get_ls returns the address of the spe local store in system memory.
 *
 * @param speid spe store structure
 * 
 * \returns address of local store.
 */
void
*spe_get_ls (speid_t speid)
{
#ifdef SPE_NO_MMAP
	errno = EFAULT;
	return NULL;
#else
	struct thread_store *thread_store = speid;

	return thread_store->mem_mmap_base;
#endif
}

/**
 * \brief spe_get_priority - returns scheduling priority for given thread group
 * @param gid spe group store structure
 * 
 * \returns scheduling priority.
 */
int 
spe_get_priority(spe_gid_t gid)
{
	struct group_store *group_store = gid;
	
	return group_store->priority;
}

/**
 * \brief spe_set_priority - sets scheduling priority for given thread group
 * @param gid spe group store structure
 * @param priority priority to set.
 * 
 * \retval zero on success
 * \retval -1 on error.
 */
int 
spe_set_priority(spe_gid_t gid, int priority)
{
	struct group_store *group_store = gid;
	group_store->priority=priority;

	/*ToDo: Sanity check for priority versus policy*/
	/*ToDo: Actually change the priority for all 
	        members of the group. */

	return 0;
}

/**
 * \brief spe_get_policy - returns scheduling policy for given thread group
 * @param gid spe store structure
 * 
 * \returns scheduling policy
 */
int 
spe_get_policy(spe_gid_t gid)
{
	struct group_store *group_store = gid;
	return group_store->policy;
}

/**
 * \brief spe_get_group - returns group store structure a spe thread belongs to
 * @param speid spe store structure
 * 
 * \returns group store structure the thread belongs to
 */
spe_gid_t
spe_get_group (speid_t speid)
{
	struct thread_store *thread_store = speid;
	return thread_store->group_id;
}

/**
 * \brief spe_get_ps - returns pointer to problem state for mfc wrapper calls (RFU)
 *
 * spe_get_ps returns a pointer to the required input for _mfc calls. In this 
 * implementation it passes the speid back.
 * 
 * @param speid spe store structure
 * 
 * \returns thread store structure.
 */
void
*spe_get_ps (speid_t speid)
{
	return speid;
}

/**
 * \brief spe_group_defaults - sets default thread group attributes.
 * @param policy scheduling class for spe threads
 * @param priority spe groups scheduling priority
 * @param spe_events RFU
 * 
 * \retval zero on success
 * \retval -1 on failure.
 */
int
spe_group_defaults(int policy, int priority, int spe_events)
{
	/*ToDo: Sanity check for priority versus policy*/
	
	default_group.policy = policy;
	default_group.priority = priority;
	default_group.use_events = spe_events;
	
	return 0;
}

/*
 * mfc.h direct call-ins
 *
 * */

/**
 * \brief _put_mfc_outbox - MFC wrapper function, documented in mfc.h.
 */
int _put_mfc_outbox (void *spe_ps_addr ,unsigned int data)
{
	struct thread_store *thread_store = spe_ps_addr;
	int rc;
	
	rc = write(thread_store->fd_wbox, &data, 4);
	if (rc == 4)
		rc = 0;

	return rc;

}

/**
 * \brief  _stat_mfc_outbox - MFC wrapper function, documented in mfc.h.
 */
int _stat_mfc_outbox (void *spe_ps_addr)
{
	struct thread_store *thread_store = spe_ps_addr;
	int rc, ret;

	rc = read(thread_store->fd_wbox_stat, &ret, 4);
	if (rc != 4)
		ret = -1;

	return ret;

}

/**
 * \brief  _get_mfc_inbox - MFC wrapper function, documented in mfc.h.
 */
int _get_mfc_inbox (void *spe_ps_addr)
{
        struct thread_store *thread_store = spe_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_mbox, &ret, 4);
        if (rc != 4)
	        ret = -1;

        return ret;

}

/**
 * \brief _stat_mfc_inbox - MFC wrapper function, documented in mfc.h.
 */
int _stat_mfc_inbox (void *spe_ps_addr)
{
	struct thread_store *thread_store = spe_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_mbox_stat, &ret, 4);
        if (rc != 4)
                ret = -1;

        return ret;
}

/**
 * \brief  _stat_mfc_inbox_intr - MFC wrapper function, documented in mfc.h.
 */
int _stat_mfc_inbox_intr (void *spe_ps_addr)
{
        struct thread_store *thread_store = spe_ps_addr;
        int rc, ret;

        rc = read(thread_store->fd_ibox_stat, &ret, 4);
        if (rc != 4)
                ret = -1;

        return ret;

}

/**
 * \brief _send_mfc_sig - MFC wrapper function, documented in mfc.h.
 */
int _send_mfc_sig (void *spe_ps_addr, unsigned int signal_reg, unsigned int data)
{
	struct thread_store *thread_store = spe_ps_addr;
	int rc;

	rc = write(thread_store->fd_sig1, &data, 4);
	if (rc == 4)
		rc = 0;

	return rc;
}

/*@}*/

